class Table(object):
    def __init__(self, name, columns):
        self.name = name
        self.columns = columns

    def __repr__(self):
        columns = [str(column) for column in self.columns]
        return '<Table name={name}>\n  {columns}\n</Table>'.format(
            name=self.name,
            columns='\n  '.join(columns),
        )



from .execution_plan import ExecutionPlan
from .clear_table_execution_step import ClearTableExecutionStep
from .clear_table_column_execution_step import ClearTableColumnExecutionStep

class ExecutionPlanner(object):
    '''
    Turns a definition into an execution plan.
    '''
    def create(self, definition):
        plan=ExecutionPlan()

        for table in definition.tables:
            sub_plan = ExecutionPlan(table.name)

            if table.action == 'clear':
                sub_plan.add_step(ClearTableExecutionStep(table.name))
            elif len(table.fields) > 0:
                for field in table.fields:
                    if field.action == 'clear':
                        sub_plan.add_step(ClearTableColumnExecutionStep(table.name, field.name))

            if len(sub_plan.execution_steps):
                plan.add_step(sub_plan)

        return plan


import argparse

from .db import ProviderFactory
from .definition import DefinitionFileReader, DefinitionBuilder
from .execution import ExecutionPlanner

class CLI(object):
    def __init__(self):
        self.parser = self._make_parser()

    def _make_parser(self):
        parser = argparse.ArgumentParser(
            description='Optional app description'
        )

        parser.add_argument(
            '--filepath',
            type=str,
            help='The path to a Scrubby definition file (.scrub).'
        )

        parser.add_argument(
            '--database',
            type=str,
            help='Database connection string.'
        )

        parser.add_argument(
            '--explain',
            type=bool,
            nargs='?',
            help='Explain the execution plan.'
        )

        parser.add_argument(
            '--exclude',
            type=str,
            nargs='?',
            help='Tags in definition to exclude.'
        )

        parser.add_argument(
            '--include',
            type=str,
            nargs='?',
            help='Tags in definition to include.'
        )

        return parser

    def _action_run(self, db, definition, options):
        execution_planner = ExecutionPlanner()

        execution_plan = execution_planner.create(definition)

        if options['explain']:
            print(execution_plan.explain())
        else:
            execution_plan.run(db)

    def run(self):
        db = None
        options = {}

        args = self.parser.parse_args()

        definition_reader = DefinitionFileReader()
        definition = definition_reader.read(args.filepath)

        provider_factory = ProviderFactory()
        db = provider_factory.create(args.database)

        with db:
            db_tables = db.get_tables()
            definition_builder = DefinitionBuilder(db_tables)
            definition = definition_builder.build(definition)

            if args.exclude:
                options['exclude'] = args.exclude

            if args.include:
                options['include'] = args.include

            options['explain'] = bool(args.explain)

            self._action_run(db, definition, options)

def main():
    CLI().run()


import psycopg2

from .provider import Provider
from ..entity import Table, Column

class PostgresProvider(Provider):
    def __init__(self, host, username, password, name):
        self.host = host
        self.username = username
        self.password = password
        self.name = name
        self._connection = None
        self._cursor = None

    def connect(self):
        self._connection = psycopg2.connect("dbname='{name}' user='{username}' host='{host}' password='{password}'".format(
            host=self.host,
            username=self.username,
            password=self.password,
            name=self.name,
        ))

        self._cursor = self._connection.cursor()

    def close(self):
        self._cursor.close()
        self._connection.close()

    def get_table_columns(self, table_name):
        columns = []

        with self._connection.cursor() as column_cursor:
            column_cursor.execute('SELECT column_name, is_nullable, data_type, character_maximum_length, column_default FROM information_schema.columns WHERE table_catalog=\'{name}\' AND table_name=\'{table}\''.format(
                name=self.name,
                table=table_name,
            ))

            for column_name, is_nullable, data_type, character_maximum_length, column_default in column_cursor.fetchall():
                columns.append(Column(
                    name=column_name,
                    type=data_type,
                    default_value=column_default,
                    is_primary=None,
                ))

        return columns

    def get_tables(self):
        tables = []

        def get_table_primary_key(table_name):
            query="SELECT pg_attribute.attname, format_type(pg_attribute.atttypid, pg_attribute.atttypmod) FROM pg_index, pg_class, pg_attribute, pg_namespace WHERE pg_class.oid = 'decisioning_modelresult'::regclass AND indrelid = pg_class.oid AND nspname = 'public' AND pg_class.relnamespace = pg_namespace.oid AND pg_attribute.attrelid = pg_class.oid AND pg_attribute.attnum = any(pg_index.indkey) AND indisprimary"

        with self._connection.cursor() as table_cursor:
            table_cursor.execute('SELECT relname FROM pg_class WHERE relkind=\'r\' AND relname !~ \'^(pg_|sql_)\';')

            for table in table_cursor.fetchall():
                table_name = table[0]

                columns = self.get_table_columns(table_name)

                tables.append(Table(table_name, columns))

        return tables

    def execute(self, query, **kwargs):
        formatted_query = query.format(**kwargs)
        #print('Executing query %s' % formatted_query)
        return self._cursor.execute(formatted_query)

    def clear_table(self, table_name):
        return self.execute(
            'TRUNCATE TABLE {table};',
            table=table_name,
        )

    def clear_table_column(self, table_name, column_name):
        return self.execute(
            'UPDATE {table} SET "{column}"=\'\';',
            table=table_name,
            column=column_name,
        )


from .provider import Provider

class NullProvider(Provider):
    def connect(self):
        pass

    def close(self):
        pass

    def get_table_columns(self, table_name):
        return []

    def get_tables(self):
        return []

    def execute(self, query, **kwargs):
        return None

    def clear_table(self, table_name):
        pass

    def clear_table_column(self, table_name, column_name):
        pass


from .null_provider import NullProvider
from .postgres_provider import PostgresProvider

class ProviderFactory(object):
    providers = {
        'postgres': PostgresProvider,
    }

    def create(self, connection_string):
        providers = ProviderFactory.providers

        if not connection_string:
            return NullProvider()

        options = connection_string.split(';')
        provider = options.pop(0)

        if not provider in providers:
            raise Exception('Provider {provider} is not supported.'.format(
                provider=provider,
            ))

        return providers[provider](*options)


class DefinitionBuilder(object):
    def __init__(self, db_tables):
        self.db_tables = db_tables

    def build(self, definition):
        errors = []
        valid_tables = {}

        for table in self.db_tables:
            valid_tables[table.name] = None

        for index, table in enumerate(definition.tables):
            if table.name == '*':
                # Iterate all fields and
                del definition.tables[index]
            elif not table.name in valid_tables:
                errors.append('Table {table} does not exist.'.format(
                    table=table.name,
                ))

        if errors:
            print('Found errors when parsing file:')

            for error in errors:
                print(' - ' + error)

            exit()

        return definition


from .table_field_definition import TableFieldDefinition

class TableDefinition(object):
    def __init__(self, name, action, fields):
        self.name = name
        self.action = action
        self.fields = fields

    @staticmethod
    def from_yaml(data):
        table, options = data

        fields = []

        for field in options.get('fields', {}).iteritems():
            fields.append(TableFieldDefinition.from_yaml(field))

        return TableDefinition(
            name=table,
            action=options.get('action', None),
            fields=fields,
        )

    def __repr__(self):
        return '<Table name={name} action={action}>\n{fields}\n</Table>'.format(
            name=self.name,
            action=self.action,
            fields='\n'.join(['  ' + str(field) for field in self.fields]),
        )


'''
Handles packaging, distribution, and testing.
'''

from sys import exit
from subprocess import call
from setuptools import Command, find_packages, setup

class BaseCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

class TestCommand(BaseCommand):
    description = 'run tests'

    def run(self):
        exit(call(['py.test', '--quiet']))

class ReleaseCommand(BaseCommand):
    description = 'Cut a new PyPI release.'

    def run(self):
        call(['rm', '-rf', 'build', 'dist'])
        ret = call(['python', 'setup.py', 'sdist', 'bdist_wheel', '--universal', 'upload'])
        exit(ret)

setup(
    # Basic package information.
    name = 'scrubby',
    version = '0.1.8',
    packages = find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),

    # Packaging options.
    zip_safe = False,
    include_package_data = True,
    platforms='any',
    entry_points={
        'console_scripts': [
            'scrubby = scrubby.cli:main',
        ],
    },

    # Package dependencies.
    install_requires = ['pyyaml==3.12'],
    extras_require = {
        'test': ['pytest', 'pytest-cov'],
    },
    cmdclass = {
        'test': TestCommand,
        'release': ReleaseCommand,
    },

    # Metadata for PyPI.
    author = 'Robin Orheden',
    author_email = 'orhedenr@gmail.com',
    license = 'LICENSE.md',
    url = 'http://github.com/typerandom/scrubby',
    keywords = 'data scrubbing',
    description = 'Let Scrubby clean your datasource.',
    long_description = open('README.md').read()

)


from .execution_step import ExecutionStep

class ClearTableExecutionStep(ExecutionStep):
    def __init__(self, table_name):
        self.table_name = table_name

    def run(self, db):
        return db.clear_table(self.table_name)

    def explain(self):
        return 'Clear all rows on table {table}.'.format(
            table=self.table_name,
        )